/*
 *
 * Copyright (c) 2004-2005 Guido Falsi <mad@madpilot.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: DoubleStrDouble.c,v 1.7 2005/03/06 12:15:30 mad Exp $
 *
 */

/* 
 *
 * The printDouble function is taken from the public domain. Written
 * by Fred Bayer, can be found at http://www.lispme.de/lispme/gcc/tech.html.
 *
 * The version used here was modified by Warren Young <tangent@cyberport.com>
 * and adopted for palm usage by Ben Combee <bcombee@metrowerks.com>
 *
 * This version was taken from: http://tangentsoft.net/palmfaq/src/pdouble.c
 *
 * Any further modifications that can be found in this file are
 * from Guido Falsi.
 *
 * Specifically this function was modified by Guido Falsi to respect
 * the PalmOS prefNumberFormat system preference.
 *
 */

#include <PalmOS.h>
#include "miscdefs.h"
#include "PalmSS.h"

static char	*NPtr = NULLPTRERR;

#define NUM_DIGITS   5
#define MIN_FLOAT    5
#define ROUND_FACTOR 1.000005 /* NUM_DIGITS zeros */

static double pow1[] = {
    1e256, 1e128, 1e064,
    1e032, 1e016, 1e008,
    1e004, 1e002, 1e001
};

static double pow2[] = {
    1e-256, 1e-128, 1e-064,
    1e-032, 1e-016, 1e-008,
    1e-004, 1e-002, 1e-001
};

void printDouble(double x, Char *s)
{
    FlpCompDouble fcd;
    short e, e1, i;
    double *pd, *pd1;
    Char thSep, decSep, *start, sign = '\0';
    short dec = 0;

    /* save string start location to localize at the end of the function. */
    start = s;

    /*------------------------------------------------------------------*/
    /* Round to desired precision                                       */
    /* (this doesn't always provide a correct last digit!)              */
    /*------------------------------------------------------------------*/
    x = x * ROUND_FACTOR;

    /*------------------------------------------------------------------*/
    /* check for NAN, +INF, -INF, 0                                     */
    /*------------------------------------------------------------------*/
    fcd.d = x;
    if ((fcd.ul[0] & 0x7ff00000) == 0x7ff00000)
        if (fcd.fdb.manH == 0 && fcd.fdb.manL == 0)
            if (fcd.fdb.sign)
                StrCopy(s, "[-inf]");
            else
                StrCopy(s, "[inf]");
        else
            StrCopy(s, "[nan]");
    else if (FlpIsZero(fcd))
        StrCopy(s, "0");
    else {
        /*----------------------------------------------------------------*/
        /* Make positive and store sign                                   */
        /*----------------------------------------------------------------*/
        if (FlpGetSign(fcd)) {
            *s++ = '-';
            FlpSetPositive(fcd);
        }

        if ((unsigned) fcd.fdb.exp < 0x3ff) {   /* meaning x < 1.0 */
            /*--------------------------------------------------------------*/
            /* Build negative exponent                                      */
            /*--------------------------------------------------------------*/
            for (e = 1, e1 = 256, pd = pow1, pd1 = pow2; e1;
                 e1 >>= 1, ++pd, ++pd1)
                if (*pd1 > fcd.d) {
                    e += e1;
                    fcd.d = fcd.d * *pd;
                }
            fcd.d = fcd.d * 10.0;

            /*--------------------------------------------------------------*/
            /* Only print big exponents                                     */
            /*--------------------------------------------------------------*/
            if (e <= MIN_FLOAT) {
                *s++ = '0';
                *s++ = '.';
                dec = -1;
                while (--e)
                    *s++ = '0';
            }
            else
                sign = '-';
        }
        else {
            /*--------------------------------------------------------------*/
            /* Build positive exponent                                      */
            /*--------------------------------------------------------------*/
            for (e = 0, e1 = 256, pd = pow1, pd1 = pow2; e1;
                 e1 >>= 1, ++pd, ++pd1)
                if (*pd <= fcd.d) {
                    e += e1;
                    fcd.d = fcd.d * *pd1;
                }
            if (e < NUM_DIGITS)
                dec = e;
            else
                sign = '+';
        }

        /*----------------------------------------------------------------*/
        /* Extract decimal digits of mantissa                             */
        /*----------------------------------------------------------------*/
        for (i = 0; i < NUM_DIGITS; ++i, --dec) {
            Int32 d = fcd.d;
            *s++ = d + '0';
            if (!dec)
                *s++ = '.';
            fcd.d = fcd.d - (double)d;
            fcd.d = fcd.d * 10.0;
        }

        /*----------------------------------------------------------------*/
        /* Remove trailing zeros and decimal point                        */
        /*----------------------------------------------------------------*/
	/* I want the number to appear with four decimals anyway. */
/*        while (s[-1] == '0')
            *--s = '\0';
        if (s[-1] == '.')
            *--s = '\0'; */

        /*----------------------------------------------------------------*/
        /* Append exponent                                                */
        /*----------------------------------------------------------------*/
        if (sign) {
            *s++ = 'e';
            *s++ = sign;
            StrIToA(s, e);
        }
        else
            *s = '\0';
    }
    /* Convert the number to localized format. */
    LocGetNumberSeparators(numberFormat, &thSep, &decSep);
    StrLocalizeNumber(start, thSep, decSep);
}

double my_atof(const char *pc)
{
	FlpCompDouble fcd;
	Char thSep, decSep, *s;

	/* Convert the number from the localized format. */
	s = MemPtrNew(BUFSIZE);
	ErrFatalDisplayIf(s == NULL, NPtr);
	StrNCopy(s, pc, BUFSIZE-1);
	s[BUFSIZE-1] = chrNull;
	LocGetNumberSeparators(numberFormat, &thSep, &decSep);
	StrDelocalizeNumber(s, thSep, decSep);

	FlpBufferAToF(&fcd.fd, s);

	MemPtrFree(s);

	return fcd.d;
}
