/*
 *
 * Copyright (c) 2004-2005 Guido Falsi <mad@madpilot.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: PSSdb.c,v 1.22.2.1 2005/03/30 08:17:45 mad Exp $
 *
 */

#include <PalmOS.h>

#include "miscdefs.h"
#include "PalmSS.h"
#include "MatchList.h"
#include "PalmSSRsc.h"
#include "PSSdb.h"

/* functions that need to be defined beforehand */
static Err StageDBDeleteRecord(UInt16 index);
static Err CompDBDeleteRecord(UInt16 index);
static Err ScoreDBCreateScore(UInt16 match, UInt32 stageuid, UInt32 compuid);
static Err ScoreDBDeleteScore(UInt16 match, Int16 compsel, Int16 stagesel);

/* keep it in variable so we don't have to access DM every time */
static UInt16 numMatches = 0;
static UInt16 numStagesInM = 0;
static UInt16 numCompsInM = 0;

/* Packed record structs typedefs */

typedef struct {
	DateType date;
	UInt32 StgRef;
	UInt32 CptRef;
	UInt16 StgCnt;
	UInt16 CptCnt;
	Char placename;
} MatchPkt;

typedef MatchPkt * MatchPktPtr;

typedef struct {
	UInt16 num;
	UInt8 papers;
	UInt8 irons;
	UInt8 disapp;
	UInt8 penalties;
	UInt32 MtcRef;
	UInt32 prev;
	UInt32 next;
	Char name;
} StagePkt;

typedef StagePkt * StagePktPtr;

typedef struct {
	UInt16 num;
	UInt16 division:2;
	UInt16 PF:1;
	UInt16 DQ:1;
	UInt32 MtcRef;
	UInt32 ScrRef;
	UInt32 prev;
	UInt32 next;
	Char name;
} CompPkt;

typedef CompPkt * CompPktPtr;

/*
 * The Score structure is a fixed size one (28 bytes?) so no
 * need for a packed struct
 */

typedef Score ScorePkt;
typedef ScorePtr ScorePktPtr;

/* Open databses references */

static DmOpenRef                MatchDB;
static DmOpenRef                StageDB;
static DmOpenRef                CompDB;
static DmOpenRef                ScoreDB;

static Err SetDirty(DmOpenRef dbref, UInt16 *index)
{
	UInt16 rattr;
	Err err;

	rattr = true;

	err = DmRecordInfo(dbref, *index, &rattr, NULL, NULL);
	if(err) return err;
	rattr = rattr | dmRecAttrDirty;
	err = DmSetRecordInfo(dbref, *index, &rattr, NULL);
	if(err) return err;

	return errNone;
}

static Err GetDB(DmOpenRef *dbPP, const Char *DBName, UInt32 DBType, UInt16 mode)
{
	Err err;
	DmOpenRef dbP;
	LocalID dbID;
	UInt16 cardNo, attrs;

	*dbPP = NULL;

	dbP = DmOpenDatabaseByTypeCreator(DBType, appID, mode);
	if(!dbP)
	{
		err = DmCreateDatabase(DBcardNo, DBName, appID, DBType, false);
		if(err) return err;

		dbP = DmOpenDatabaseByTypeCreator(DBType, appID, mode);
		if(!dbP) return DmGetLastErr();

		DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL);
		if(err)
		{
			DmCloseDatabase(dbP);
			return err;
		}
		DmDatabaseInfo(cardNo, dbID, NULL, &attrs, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		if(err)
		{
			DmCloseDatabase(dbP);
			return err;
		}
		attrs |= dmHdrAttrBackup;
		DmSetDatabaseInfo(cardNo, dbID, NULL, &attrs, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
		if(err)
		{
			DmCloseDatabase(dbP);
			return err;
		}
	}

	*dbPP = dbP;
	return errNone;
}

Err OpenDBs()
{
	Err err;

	err = GetDB(&MatchDB, MatchDBName, MatchDBType, dmModeReadWrite);
	if(err) return err;
	err = GetDB(&StageDB, StageDBName, StageDBType, dmModeReadWrite);
	if(err) return err;
	err = GetDB(&CompDB, CompDBName, CompDBType, dmModeReadWrite);
	if(err) return err;
	err = GetDB(&ScoreDB, ScoreDBName, ScoreDBType, dmModeReadWrite);
	if(err) return err;

	return errNone;
}

void CloseDBs()
{
	DmCloseDatabase(MatchDB);
	DmCloseDatabase(StageDB);
	DmCloseDatabase(CompDB);
	DmCloseDatabase(ScoreDB);
}

Err MatchDBTableDataFree(MatchTableDataPtr *DataArray, Boolean new)
{
	Int16 i;

	if(*DataArray)
	{
		if((i = MatchDBNumRecords()) > 0)
		{
			i = new ? i-- : i;
			while(i)
			{
				i--;
				MemHandleUnlock((*DataArray)[i].recordH);
			}

			MemPtrFree(*DataArray);
		}
	}

	return errNone;
}

Err MatchDBTableDataQuery(MatchTableDataPtr *DataArray)
{
	MemHandle recordH;
	MatchPktPtr record;
	UInt16 i, j;
	Err err = errNone;

	if((i = MatchDBNumRecords()) > 0)
	{
		*DataArray = MemPtrNew(sizeof(MatchTableData) * i);
		if(!*DataArray) return -1;

		for(j = 0; j < i; j++)
		{
			recordH = DmQueryRecord(MatchDB, j);
			if(!recordH) return err;

			record = MemHandleLock(recordH);
			if(!record) return 1;

			/* find the date and link it in the DataArray structure. */
			(*DataArray)[j].date = &record->date;
	
			/*
			 * Find the place string in the record and put a pointer
			 * to it in the result structure.
			 */

			(*DataArray)[j].place = &record->placename;

			/*
			 * Put an handle to the record in the structure,
			 * so we can free it cleanly later.
			 */

			(*DataArray)[j].recordH = recordH;
		}
	} else {
		*DataArray = NULL;
	}

	return errNone;
}

static UInt32 MatchDBPackedSize(MatchPtr record)
{
	UInt32 size;

	size = sizeof(DateType) + sizeof(UInt16) * 2 + sizeof(UInt32) * 2;
	size += StrLen(record->place) + 1;
	size += StrLen(record->name) + 1;

	return size;
}

static void MatchDBPackRecord(MatchPtr unpacked, MatchPktPtr packed)
{
	UInt32 offset = 0;

	DmWrite(packed, offset, &unpacked->date, sizeof(DateType) + sizeof(UInt16) * 2 + sizeof(UInt32) * 2);
	offset += sizeof(DateType) + sizeof(UInt16) * 2 + sizeof(UInt32) * 2;
	DmWrite(packed, offset, unpacked->place, StrLen(unpacked->place) + 1);
	offset += StrLen(unpacked->place) + 1;
	DmWrite(packed, offset, unpacked->name, StrLen(unpacked->name) + 1);
}

static void MatchDBUnpackRecord(MatchPktPtr packed, MatchPtr unpacked)
{
	Char *p;

	unpacked->date = packed->date;
	unpacked->StgRef = packed->StgRef;
	unpacked->CptRef = packed->CptRef;
	unpacked->StgCnt = packed->StgCnt;
	unpacked->CptCnt = packed->CptCnt;
	p = &packed->placename;
	StrCopy(unpacked->place, p);
	p += StrLen(p) + 1;
	StrCopy(unpacked->name, p);
}

Err MatchDBCreateMatch(MatchPtr unpacked, UInt16 *index)
{
	MemHandle recordH;
	MatchPktPtr recordP;
	UInt16 newIndex;
	UInt32 size;

	size = MatchDBPackedSize(unpacked);
	newIndex = dmMaxRecordIndex;
	recordH = DmNewRecord(MatchDB, &newIndex, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	MatchDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	*index = newIndex;

	DmReleaseRecord(MatchDB, newIndex, true);

	/* Keep the counter updated */
	numMatches++;

	return errNone;
}

Err MatchDBDeleteMatch(UInt16 index)
{
	Err err;
	MatchPktPtr recordP;
	MemHandle recordH;
	UInt32 stageuid, compuid;
	UInt16 stageindex = NoRecord, compindex = NoRecord;

	/* Remove linked stages. */

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH) return DmGetLastErr();
	recordP = MemHandleLock(recordH);
	if(!recordP) return -1;
	stageuid = recordP->StgRef;
	compuid = recordP->CptRef;

	while(stageuid != NoRecord)
	{	
		err = DmFindRecordByID(StageDB, stageuid, &stageindex);
		if(err) return err;
		err = StageDBDeleteRecord(stageindex);
		if(err) return err;

		/* Get next one if any... */
		stageuid = recordP->StgRef;
	}

	while(compuid != NoRecord)
	{
		err = DmFindRecordByID(CompDB, compuid, &compindex);
		if(err) return err;
		err = CompDBDeleteRecord(compindex);
		if(err) return err;

		/* Get next one if any... */
		compuid = recordP->CptRef;
	}

	MemHandleUnlock(recordH);

	/*
	 * Code to remove linked scores will be added when code to
	 * manage scores will be ready.
	 */

	err = DmRemoveRecord(MatchDB, index);
	if(err) return err;

	numMatches--;

	return errNone;
}

Err MatchDBUpdateMatch(MatchPtr unpacked, UInt16 *index)
{
	MemHandle recordH, oldH;
	MatchPktPtr recordP;
	UInt32 size;
	Err err;

	oldH = (MemHandle)true;

	size = MatchDBPackedSize(unpacked);
	recordH = DmNewHandle(MatchDB, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

        MatchDBPackRecord(unpacked, recordP);
        MemHandleUnlock(recordH);

	err = DmAttachRecord(MatchDB, index, recordH, &oldH);
	if(err) return err;
	MemHandleFree(oldH);

	SetDirty(MatchDB, index);

	return errNone;
}

Err MatchDBGetMatch(UInt16 index, MatchPtr record)
{
	MemHandle recordH;
	MatchPktPtr packed;

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH)
	{
		record = NULL;
		return DmGetLastErr();
	}

	packed = (MatchPktPtr)MemHandleLock(recordH);
	if(!packed)
	{
		record = NULL;
		return 1;
	}

	MatchDBUnpackRecord(packed, record);
	MemHandleUnlock(recordH);

	return errNone;
}

Err StageDBTableDataFree(UInt16 mindex, StageTableDataPtr *DataArray, Boolean new)
{
	Int16 i;

	if(*DataArray)
	{
		if((i = StageInMatchDBNumRecords(mindex)) > 0)
		{
			i = new ? i-- : i;
			while(i)
			{
				i--;
				MemHandleUnlock((*DataArray)[i].recordH);
			}

			MemPtrFree(*DataArray);
		}
	}

	return errNone;
}

Err StageDBTableDataQuery(UInt16 mindex, StageTableDataPtr *DataArray)
{
	MemHandle recordH;
	StagePktPtr srecord = NULL; /* make gcc quiet */
	MatchPktPtr mrecord;
	UInt16 sindex, i;
	UInt32 stageuid;
	Err err;

	recordH = DmQueryRecord(MatchDB, mindex);
	if(!recordH)
	{
		err = DmGetLastErr();
		return err;
	}

	mrecord = MemHandleLock(recordH);
	if(!mrecord) return 1;

	stageuid = mrecord->StgRef;

	MemHandleUnlock(recordH);

	if((i = StageInMatchDBNumRecords(mindex)) > 0)
	{

		*DataArray = MemPtrNew(sizeof(StageTableData) * i);
		if(!*DataArray) return -1;

		i = 0;
		do
		{
			err = DmFindRecordByID(StageDB, stageuid, &sindex);
			if(err) return err;

			recordH = DmQueryRecord(StageDB, sindex);
			if(!recordH)
			{
				err = DmGetLastErr();
				return err;
			}

			srecord = MemHandleLock(recordH);
			if(!srecord) return 1;

			/* Calculate points in stage and return for listing */
			(*DataArray)[i].mpts = ((srecord->papers + srecord->disapp) * 2 + srecord->irons) * 5;
	
			/*
			 * Find the name string in the record and put a pointer
			 * to it in the result pointer.
			 */
			(*DataArray)[i].name = &(srecord->name);

			(*DataArray)[i].recordH = recordH;

			stageuid = srecord->next;
			i++;
		} while(stageuid != NoRecord);
	} else {
		*DataArray = NULL;
	}

	return errNone;
}

Int16 StageDBGetIndexInList(UInt16 index)
{
	MemHandle recordH;
	StagePktPtr recordP;
	UInt32 prevuid;
	Int16 ret;
	Err err;

	if(index == NoRecord) return NoRecord;

	recordH = DmQueryRecord(StageDB, index);
	if(!recordH) return NoRecord;
	recordP = MemHandleLock(recordH);
	if(!recordP) return NoRecord;

	prevuid = recordP->prev;

	MemHandleUnlock(recordH);

	for(ret = 0; prevuid != NoRecord; ret++)
	{
		err = DmFindRecordByID(StageDB, prevuid, &index);
		if(err) return NoRecord;

		recordH = DmQueryRecord(StageDB, index);
		if(!recordH) return NoRecord;
		recordP = MemHandleLock(recordH);
		if(!recordP) return NoRecord;

		prevuid = recordP->prev;

		MemHandleUnlock(recordH);
	}

	return ret;
}

static UInt16 StageDBGetRecordInList(UInt16 mindex, Int16 selection)
{
	MemHandle recordH;
	StagePktPtr srecord = NULL; /* make gcc quiet */
	MatchPktPtr mrecord;
	UInt16 sindex;
	UInt32 stageuid;
	Err err;

	if(selection == NoRecord)
		return NoRecord;

	recordH = DmQueryRecord(MatchDB, mindex);
	if(!recordH) return NoRecord;
	mrecord = MemHandleLock(recordH);
	if(!mrecord) return NoRecord;

	stageuid = mrecord->StgRef;

	MemHandleUnlock(recordH);

	for(;selection > 0; selection--)
	{
		err = DmFindRecordByID(StageDB, stageuid, &sindex);
		if(err) return NoRecord;

		recordH = DmQueryRecord(StageDB, sindex);
		if(!recordH) return NoRecord;
		srecord = MemHandleLock(recordH);
		if(!srecord) return NoRecord;

		stageuid = srecord->next;

		MemHandleUnlock(recordH);
	}

	err = DmFindRecordByID(StageDB, stageuid, &sindex);
	if(err) return NoRecord;

	return sindex;
}

static UInt32 StageDBPackedSize(StagePtr record)
{
	UInt32 size;

	size = sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32) * 3;
	size += StrLen(record->name) + 1;

	return size;
}

static void StageDBPackRecord(StagePtr unpacked, StagePktPtr packed)
{
	UInt32 offset = 0;

	DmWrite(packed, offset, &unpacked->num, sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32) * 3);
	offset += sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32) * 3;
	DmWrite(packed, offset, unpacked->name, StrLen(unpacked->name) + 1);
}

static void StageDBUnpackRecord(StagePktPtr packed, StagePtr unpacked)
{
	unpacked->num = packed->num;
	unpacked->papers = packed->papers;
	unpacked->irons = packed->irons;
	unpacked->disapp = packed->disapp;
	unpacked->penalties = packed->penalties;
	unpacked->MtcRef = packed->MtcRef;
	unpacked->prev = packed->prev;
	unpacked->next = packed->next;
	StrCopy(unpacked->name, &packed->name);
}

static Err MatchSetStageData(UInt16 matchi, UInt32 stageuid)
{
	MemHandle recordH, recordHH;
	MatchPktPtr pktmatch;
	StagePktPtr pktstage, pktstagee;
	UInt16 sindex;
	UInt32 i;
	Err err;
	
	recordH = DmGetRecord(MatchDB, matchi);
	if(!recordH) return DmGetLastErr();

	pktmatch = MemHandleLock(recordH);
	if(!pktmatch) return 1;

	DmWrite(pktmatch, sizeof(DateType) + sizeof(UInt32) * 2, &numStagesInM, sizeof(UInt16));

	if(stageuid != NoRecord)
	{
		if((i = pktmatch->StgRef) != NoRecord)
		{
			do
			{
				/*
				 * unlock record from previous iteration.
				 */
				if(recordH) MemHandleUnlock(recordH);

				err = DmFindRecordByID(StageDB, i, &sindex);

				recordH = DmQueryRecord(StageDB, sindex);
				if(!recordH) return DmGetLastErr();

				pktstage = MemHandleLock(recordH);
				if(!pktstage) return -1;

				if(pktstage->next == NoRecord)
				{
					recordHH = DmGetRecord(StageDB, sindex);
					if(!recordHH) return DmGetLastErr();

					pktstagee = MemHandleLock(recordHH);
					if(!pktstagee) return -1;

					DmWrite(pktstagee, sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32) * 2, &stageuid, sizeof(UInt32));
					MemHandleUnlock(recordHH);
					DmReleaseRecord(StageDB, sindex, true);

					break;
				} else
					i = pktstage->next;
			} while (true);
		} else {
			DmWrite(pktmatch, sizeof(DateType), &stageuid, sizeof(UInt32));
		}
	}

	MemHandleUnlock(recordH);
	DmReleaseRecord(MatchDB, matchi, true);

	return errNone;
}

Err StageDBCreateStage(StagePtr unpacked, UInt16 match, Int16 *listindex)
{
	MemHandle recordH;
	StagePktPtr recordP;
	UInt16 newIndex;
	UInt32 size, stageuid;
	Err err;

	size = StageDBPackedSize(unpacked);
	newIndex = dmMaxRecordIndex;
	recordH = DmNewRecord(StageDB, &newIndex, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	StageDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	err = DmReleaseRecord(StageDB, newIndex, true);
	if(err) return err;

	*listindex = StageDBGetIndexInList(newIndex);

	stageuid = true;
	err = DmRecordInfo(StageDB, newIndex, NULL, &stageuid, NULL);
	if(err) return err;

	err = ScoreDBCreateScore(match, stageuid, false);
	if(err) return err;

	/* Keep the counter updated */
	numStagesInM++;

	/* add the stage link to the match record */
	MatchSetStageData(match, stageuid);

	return errNone;
}

Err StageDBDeleteStage(UInt16 match, Int16 listindex)
{
	return StageDBDeleteRecord(StageDBGetRecordInList(match, listindex));
}

static Err StageDBDeleteRecord(UInt16 index)
{
	Err err;
	MemHandle recordH;
	StagePktPtr stage;
	MatchPktPtr match;
	UInt32 nextuid, prevuid, matchuid;
	UInt16 next = NoRecord, prev = NoRecord, matchindex = NoRecord;
	Int16 i;

	/* first get the record */
	recordH = DmQueryRecord(StageDB, index);
	if(!recordH) return DmGetLastErr();

	stage = MemHandleLock(recordH);
	if(!stage) return 1;

	/* and save the info we need */
	nextuid = stage->next;
	prevuid = stage->prev;
	matchuid = stage->MtcRef;

	err = DmFindRecordByID(MatchDB, matchuid, &matchindex);
	if(err) return err;

	/* Delete all connected scores */
	for(i = CompInMatchDBNumRecords(matchindex) - 1; i >= 0; i--)
		ScoreDBDeleteScore(matchindex, i, index);

	MemHandleUnlock(recordH);

	/* now relink the list */
	if(prevuid != NoRecord)
	{
		err = DmFindRecordByID(StageDB, prevuid, &prev);
		if(err) return err;

		recordH = DmGetRecord(StageDB, prev);
		if(!recordH) return DmGetLastErr();

		stage = MemHandleLock(recordH);
		if(!stage) return 1;

		DmWrite(stage, sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32) * 2, &nextuid, sizeof(UInt32));
		MemHandleUnlock(recordH);
		err = DmReleaseRecord(StageDB, prev, true);
		if(err) return err;
	} else {
		recordH = DmGetRecord(MatchDB, matchindex);
		if(!recordH) return DmGetLastErr();

		match = MemHandleLock(recordH);
		if(!match) return 1;

		err = DmWrite(match, sizeof(DateType), &nextuid, sizeof(UInt32));
		if(err) return err;
		MemHandleUnlock(recordH);
		err = DmReleaseRecord(MatchDB, matchindex, true);
		if(err) return err;
	}

	if(nextuid != NoRecord)
	{
		err = DmFindRecordByID(StageDB, nextuid, &next);
		if(err) return err;

		recordH = DmGetRecord(StageDB, next);
		if(!recordH) return DmGetLastErr();

		stage = MemHandleLock(recordH);
		if(!stage) return 1;

		err = DmWrite(stage, sizeof(UInt16) + sizeof(UInt8) * 4 + sizeof(UInt32), &prevuid, sizeof(UInt32));
		if(err) return err;

		MemHandleUnlock(recordH);
		err = DmReleaseRecord(StageDB, next, true);
		if(err) return err;
	}

	/* and now really delete the record */
	err = DmRemoveRecord(StageDB, index);
	if(err) return err;

	/* and to end it, update counters. */
	numStagesInM--;
	MatchSetStageData(matchindex, NoRecord);

	return errNone;
}

Err StageDBUpdateStage(StagePtr unpacked, UInt16 match, Int16 listindex)
{
	MemHandle recordH, oldH;
	StagePktPtr recordP;
	UInt16 index;
	UInt32 size;
	Err err;

	index = StageDBGetRecordInList(match, listindex);

	oldH = (MemHandle)true;

	size = StageDBPackedSize(unpacked);
	recordH = DmNewHandle(StageDB, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	StageDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	err = DmAttachRecord(StageDB, &index, recordH, &oldH);
	if(err) return err;
	MemHandleFree(oldH);

	SetDirty(StageDB, &index);

	return errNone;
}

Err StageDBGetStage(UInt16 match, Int16 listindex, StagePtr record)
{
	MemHandle recordH;
	StagePktPtr packed;
	UInt16 index;

	index = StageDBGetRecordInList(match, listindex);

	recordH = DmQueryRecord(StageDB, index);
	if(!recordH)
	{
		record = NULL;
		return DmGetLastErr();
	}

	packed = MemHandleLock(recordH);
	if(!packed)
	{
		record = NULL;
		return 1;
	}

	StageDBUnpackRecord(packed, record);
	MemHandleUnlock(recordH);

	return errNone;
}

UInt8 StageDBStageRounds(UInt16 match, Int16 stage)
{
	UInt16 index;
	MemHandle recordH;
	StagePktPtr record;
	UInt8 ret;

	index = StageDBGetRecordInList(match, stage);

	recordH = DmQueryRecord(StageDB, index);
	if(!recordH) return 0;
	record = MemHandleLock(recordH);
	if(!record) return 0;

	ret = record->irons + record->papers * 2 + record->disapp * 2;

	MemHandleUnlock(recordH);

	return ret;
}

Int32 StageDBGetStageMpts(UInt16 match, Int16 stage)
{
	UInt16 index;
	MemHandle recordH;
	StagePktPtr record;
	Int32 ret;

	index = StageDBGetRecordInList(match, stage);

	recordH = DmQueryRecord(StageDB, index);
	if(!recordH) return 0;
	record = MemHandleLock(recordH);
	if(!record) return 0;

	ret = ((record->papers + record->disapp) * 2 + record->irons) * 5;

	MemHandleUnlock(recordH);

	return ret;
}

Err CompDBTableDataFree(UInt16 mindex, CompTableDataPtr *DataArray, Boolean new)
{
	Int16 i;

	if(*DataArray)
	{
		if((i = CompInMatchDBNumRecords(mindex)) > 0)
		{
			i = new ? i-- : i;
			while(i)
			{
				i--;
				MemHandleUnlock((*DataArray)[i].recordH);
			}

			MemPtrFree(*DataArray);
		}
	}

	return errNone;
}

Err CompDBTableDataQuery(UInt16 mindex, CompTableDataPtr *DataArray)
{
	MemHandle recordH;
	CompPktPtr crecord = NULL; /* make gcc quiet */
	MatchPktPtr mrecord;
	UInt16 cindex, i;
	UInt32 compuid;
	Err err;

	recordH = DmQueryRecord(MatchDB, mindex);
	if(!recordH)
	{
		err = DmGetLastErr();
		return err;
	}

	mrecord = MemHandleLock(recordH);
	if(!mrecord) return 1;

	compuid = mrecord->CptRef;

	MemHandleUnlock(recordH);

	if((i = CompInMatchDBNumRecords(mindex)) > 0)
	{
		*DataArray = MemPtrNew(sizeof(CompTableData) * i);
		if(!*DataArray) return -1;

		i = 0;
		do
		{

			err = DmFindRecordByID(CompDB, compuid, &cindex);
			if(err) return err;

			recordH = DmQueryRecord(CompDB, cindex);
			if(!recordH)
			{
				err = DmGetLastErr();
				return err;
			}

			crecord = MemHandleLock(recordH);
			if(!crecord) return 1;

			/*
			 * Find the name string in the record and put a pointer
			 * to it in the result pointer.
			 */
			(*DataArray)[i].name = &(crecord->name);

			/* Check if disqualified */
			(*DataArray)[i].DQ = crecord->DQ;

			(*DataArray)[i].recordH = recordH;

			compuid = crecord->next;
			i++;
		} while(compuid != NoRecord);
	} else {
		*DataArray = NULL;
	}

	return errNone;
}

Int16 CompDBGetIndexInList(UInt16 index)
{
	MemHandle recordH;
	CompPktPtr recordP;
	UInt32 prevuid;
	Int16 ret;
	Err err;

	if(index == NoRecord) return NoRecord;

	recordH = DmQueryRecord(CompDB, index);
	if(!recordH) return NoRecord;
	recordP = MemHandleLock(recordH);
	if(!recordP) return NoRecord;

	prevuid = recordP->prev;

	MemHandleUnlock(recordH);

	for(ret = 0; prevuid != NoRecord; ret++)
	{
		err = DmFindRecordByID(CompDB, prevuid, &index);
		if(err) return NoRecord;

		recordH = DmQueryRecord(CompDB, index);
		if(!recordH) return NoRecord;
		recordP = MemHandleLock(recordH);
		if(!recordP) return NoRecord;

		prevuid = recordP->prev;

		MemHandleUnlock(recordH);
	}

	return ret;
}

static UInt16 CompDBGetRecordInList(UInt16 mindex, Int16 selection)
{
	MemHandle recordH;
	CompPktPtr crecord = NULL; /* make gcc quiet */
	MatchPktPtr mrecord;
	UInt16 cindex;
	UInt32 compuid;
	Err err;

	if(selection == NoRecord)
		return NoRecord;

	recordH = DmQueryRecord(MatchDB, mindex);
	if(!recordH) return NoRecord;
	mrecord = MemHandleLock(recordH);
	if(!mrecord) return NoRecord;

	compuid = mrecord->CptRef;

	MemHandleUnlock(recordH);

	for(;selection > 0; selection--)
	{
		err = DmFindRecordByID(CompDB, compuid, &cindex);
		if(err) return NoRecord;

		recordH = DmQueryRecord(CompDB, cindex);
		if(!recordH) return NoRecord;
		crecord = MemHandleLock(recordH);
		if(!crecord) return NoRecord;

		compuid = crecord->next;

		MemHandleUnlock(recordH);
	}

	err = DmFindRecordByID(CompDB, compuid, &cindex);
	if(err) return NoRecord;

	return cindex;
}

static UInt32 CompDBPackedSize(CompPtr record)
{
	UInt32 size;

	size = sizeof(UInt16) * 2 + sizeof(UInt32) * 4;
	size += StrLen(record->name) + 1;

	return size;
}

static void CompDBPackRecord(CompPtr unpacked, CompPktPtr packed)
{
	UInt32 offset = 0;

	DmWrite(packed, offset, &unpacked->num, sizeof(UInt16) * 2 + sizeof(UInt32) * 4);
	offset += sizeof(UInt16) * 2 + sizeof(UInt32) * 4;
	DmWrite(packed, offset, unpacked->name, StrLen(unpacked->name) + 1);
}

static void CompDBUnpackRecord(CompPktPtr packed, CompPtr unpacked)
{
	unpacked->num = packed->num;
	unpacked->division = packed->division;
	unpacked->PF = packed->PF;
	unpacked->DQ = packed->DQ;
	unpacked->MtcRef = packed->MtcRef;
	unpacked->ScrRef = packed->ScrRef;
	unpacked->prev = packed->prev;
	unpacked->next = packed->next;
	StrCopy(unpacked->name, &packed->name);
}

Boolean CompDBIsCompMajor(UInt16 match, Int16 comp)
{
	UInt16 index;
	CompPktPtr record;
	MemHandle recordH;
	Boolean ret;

	index = CompDBGetRecordInList(match, comp);

	recordH = DmQueryRecord(CompDB, index);
	if(!recordH) return false;
	record = MemHandleLock(recordH);
	if(!record) return false;

	ret = (record->PF == Major);

	MemHandleUnlock(recordH);

	return ret;
}

static Err MatchSetCompData(UInt16 matchi, UInt32 compuid)
{
	MemHandle recordH, recordHH;
	MatchPktPtr pktmatch;
	CompPktPtr pktcomp, pktcompp;
	UInt16 cindex;
	UInt32 i;
	Err err;
	
	recordH = DmGetRecord(MatchDB, matchi);
	if(!recordH) return DmGetLastErr();

	pktmatch = MemHandleLock(recordH);
	if(!pktmatch) return 1;

	DmWrite(pktmatch, sizeof(DateType) + sizeof(UInt32) * 2 + sizeof(UInt16), &numCompsInM, sizeof(UInt16));

	if(compuid != NoRecord)
	{
		if((i = pktmatch->CptRef) != NoRecord)
		{
			do
			{
				/*
				 * unlock record from previous iteration.
				 */
				if(recordH) MemHandleUnlock(recordH);

				err = DmFindRecordByID(CompDB, i, &cindex);

				recordH = DmQueryRecord(CompDB, cindex);
				if(!recordH) return DmGetLastErr();

				pktcomp = MemHandleLock(recordH);
				if(!pktcomp) return -1;

				if(pktcomp->next == NoRecord)
				{
					recordHH = DmGetRecord(CompDB, cindex);
					if(!recordHH) return DmGetLastErr();

					pktcompp = MemHandleLock(recordHH);
					if(!pktcompp) return -1;

					DmWrite(pktcompp, sizeof(UInt16) * 2 + sizeof(UInt32) * 3, &compuid, sizeof(UInt32));
					MemHandleUnlock(recordHH);
					DmReleaseRecord(CompDB, cindex, true);

					break;
				} else
					i = pktcomp->next;
			} while (true);
		} else {
			DmWrite(pktmatch, sizeof(DateType) + sizeof(UInt32), &compuid, sizeof(UInt32));
		}
	}

	MemHandleUnlock(recordH);
	DmReleaseRecord(MatchDB, matchi, true);

	return errNone;
}

Err CompDBCreateComp(CompPtr unpacked, UInt16 match, Int16 *listindex)
{
	MemHandle recordH;
	CompPktPtr recordP;
	UInt16 newIndex;
	UInt32 size, compuid;
	Err err;

	size = CompDBPackedSize(unpacked);
	newIndex = dmMaxRecordIndex;
	recordH = DmNewRecord(CompDB, &newIndex, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	CompDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	err = DmReleaseRecord(CompDB, newIndex, true);
	if(err) return err;

	*listindex = CompDBGetIndexInList(newIndex);

	compuid = true;
	err = DmRecordInfo(CompDB, newIndex, NULL, &compuid, NULL);
	if(err) return err;

	err = ScoreDBCreateScore(match, false, compuid);
	if(err) return err;

	/* Keep the counter updated */
	numCompsInM++;

	/* add the stage link to the match record */
	MatchSetCompData(match, compuid);

	return errNone;
}

Err CompDBGetComp(UInt16 match, Int16 listindex, CompPtr record)
{
	MemHandle recordH;
	CompPktPtr packed;
	UInt16 index;

	index = CompDBGetRecordInList(match, listindex);

	recordH = DmQueryRecord(CompDB, index);
	if(!recordH)
	{
		record = NULL;
		return DmGetLastErr();
	}

	packed = MemHandleLock(recordH);
	if(!packed)
	{
		record = NULL;
		return 1;
	}

	CompDBUnpackRecord(packed, record);
	MemHandleUnlock(recordH);

	return errNone;
}

Err CompDBDeleteComp(UInt16 match, Int16 listindex)
{
	return CompDBDeleteRecord(CompDBGetRecordInList(match, listindex));
}

static Err CompDBDeleteRecord(UInt16 index)
{
	Err err;
	MemHandle recordH;
	CompPktPtr comp;
	MatchPktPtr match;
	UInt32 nextuid, prevuid, matchuid;
	UInt16 next = NoRecord, prev = NoRecord, matchindex = NoRecord;

	/* first get the record */
	recordH = DmQueryRecord(CompDB, index);
	if(!recordH) return DmGetLastErr();

	comp = MemHandleLock(recordH);
	if(!comp) return 1;

	/* and save the info we need */
	nextuid = comp->next;
	prevuid = comp->prev;
	matchuid = comp->MtcRef;

	err = DmFindRecordByID(MatchDB, matchuid, &matchindex);
	if(err) return err;

	/* Delete the score records connected to this competitor. */
	while(comp->ScrRef != NoRecord)
		ScoreDBDeleteScore(matchindex, CompDBGetIndexInList(index), 0);

	MemHandleUnlock(recordH);

	/* now relink the list */
	if(prevuid != NoRecord)
	{
		err = DmFindRecordByID(CompDB, prevuid, &prev);
		if(err) return err;

		recordH = DmGetRecord(CompDB, prev);
		if(!recordH) return DmGetLastErr();

		comp = MemHandleLock(recordH);
		if(!comp) return 1;

		DmWrite(comp, sizeof(UInt16) * 2 + sizeof(UInt32) * 3, &nextuid, sizeof(UInt32));
		MemHandleUnlock(recordH);
		err = DmReleaseRecord(CompDB, prev, true);
		if(err) return err;
	} else {
		recordH = DmGetRecord(MatchDB, matchindex);
		if(!recordH) return DmGetLastErr();

		match = MemHandleLock(recordH);
		if(!match) return 1;

		err = DmWrite(match, sizeof(DateType) + sizeof(UInt32), &nextuid, sizeof(UInt32));
		if(err) return err;
		MemHandleUnlock(recordH);
		err = DmReleaseRecord(MatchDB, matchindex, true);
		if(err) return err;
	}

	if(nextuid != NoRecord)
	{
		err = DmFindRecordByID(CompDB, nextuid, &next);
		if(err) return err;

		recordH = DmGetRecord(CompDB, next);
		if(!recordH) return DmGetLastErr();

		comp = MemHandleLock(recordH);
		if(!comp) return 1;

		err = DmWrite(comp, sizeof(UInt16) * 2 + sizeof(UInt32) * 2, &prevuid, sizeof(UInt32));
		if(err) return err;

		MemHandleUnlock(recordH);
		err = DmReleaseRecord(CompDB, next, true);
		if(err) return err;
	}

	/* and now really delete the record */
	err = DmRemoveRecord(CompDB, index);
	if(err) return err;

	/* and to end it, update counters. */
	numCompsInM--;
	MatchSetCompData(matchindex, NoRecord);

	return errNone;
}

Err CompDBUpdateComp(CompPtr unpacked, UInt16 match, Int16 listindex)
{
	MemHandle recordH, oldH;
	CompPktPtr recordP;
	UInt16 index;
	UInt32 size;
	Err err;

	index = CompDBGetRecordInList(match, listindex);

	oldH = (MemHandle)true;

	size = CompDBPackedSize(unpacked);
	recordH = DmNewHandle(CompDB, size);
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	CompDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	err = DmAttachRecord(CompDB, &index, recordH, &oldH);
	if(err) return err;
	MemHandleFree(oldH);

	SetDirty(CompDB, &index);

	return errNone;
}

#define ScoreDBPackedSize()	sizeof(Score)

/* static UInt32 ScoreDBPackedSize()
{
	return sizeof(Score);
} */

static void ScoreDBPackRecord(ScorePtr unpacked, ScorePktPtr packed)
{
	DmWrite(packed, 0, &unpacked->As, sizeof(Score));
}

static void ScoreDBUnpackRecord(ScorePktPtr packed, ScorePtr unpacked)
{
	unpacked->As = packed->As;
	unpacked->Cs = packed->Cs;
	unpacked->Ds = packed->Ds;
	unpacked->PTs = packed->PTs;
	unpacked->Misses = packed->Misses;
	unpacked->Procs = packed->Procs;
	unpacked->Time = packed->Time;
	unpacked->Zeroed = packed->Zeroed;
	unpacked->Scored = packed->Scored;
	unpacked->StgRef = packed->StgRef;
	unpacked->CptRef = packed->CptRef;
	unpacked->prev = packed->prev;
	unpacked->next = packed->next;
}

/*
 * NOTE: This function is a big piece oh junk, substituting it with
 * a better one is a big TODO
 */

static Err ScoreDBCreateScore(UInt16 match, UInt32 stageuid, UInt32 compuid)
{
	MemHandle mrecordH, srecordH = NULL, stgrecH = NULL, crecordH = NULL, newrecordH;
	MatchPktPtr mrecord = NULL;
	ScorePktPtr srecord = NULL, newrecord = NULL;
	CompPktPtr crecord = NULL;
	StagePktPtr stgrec = NULL;
	ScorePtr score;
	UInt16 stgidx = NoRecord, sindex = NoRecord, cindex= NoRecord, newIndex;
	UInt32 uid, stguid, scoreuid = NoRecord;
	Err err;

	if(match == NoRecord) return 1;

	mrecordH = DmQueryRecord(MatchDB, match);
	if(!mrecordH) return DmGetLastErr();

	mrecord = MemHandleLock(mrecordH);
	if(!mrecord) return 1;

	if(stageuid)
	{
		if((uid = mrecord->CptRef) != NoRecord)
		{
			do
			{
				if(crecord)
				{
					MemHandleUnlock(crecordH);
					DmReleaseRecord(CompDB, cindex, false);
				}
				err = DmFindRecordByID(CompDB, uid, &cindex);
				if(err) return err;

				crecordH = DmGetRecord(CompDB, cindex);
				if(!crecordH) return DmGetLastErr();

				crecord = MemHandleLock(crecordH);
				if(!crecord) return 1;

				compuid = uid;
				uid = crecord->ScrRef;

				do
				{
					if(srecord)
					{
						MemHandleUnlock(srecordH);
						DmReleaseRecord(ScoreDB, sindex, false);
					}
					DmFindRecordByID(ScoreDB, uid, &sindex);

					srecordH = DmGetRecord(ScoreDB, sindex);
					if(!srecordH) return DmGetLastErr();

					srecord = MemHandleLock(srecordH);
					if(!srecord) return 1;

				} while((uid = srecord->next) != NoRecord);
				/* found, now add the score record and link it */
				score = MemPtrNew(sizeof(Score));
				if(!score) return 1;

				err = DmRecordInfo(ScoreDB, sindex, NULL, &uid, NULL);
				if(err)
				{
					MemPtrFree(score);
					return err;
				}

				/* Set default values */
				score->As = 0;
				score->Cs = 0;
				score->Ds = 0;
				score->PTs = 0;
				score->Misses = 0;
				score->Procs = 0;
				score->Time = 0.0;
				score->Zeroed = 0;
				score->Scored = 0;
				score->prev = uid;
				score->next = NoRecord;
				score->StgRef = stageuid;
				score->CptRef = compuid;

				newIndex = dmMaxRecordIndex;
				newrecordH = DmNewRecord(ScoreDB, &newIndex, ScoreDBPackedSize());
				if(!newrecordH)
				{
					MemPtrFree(score);
					return DmGetLastErr();
				}
				newrecord = MemHandleLock(newrecordH);
				if(!newrecord)
				{
					MemPtrFree(score);
					return 1;
				}
				ScoreDBPackRecord(score, newrecord);
				MemPtrFree(score);

				err = DmRecordInfo(ScoreDB, newIndex, NULL, &uid, NULL);
				if(err) return err;

				DmWrite(srecord, sizeof(UInt8) * 6 + sizeof(double) + sizeof(UInt16) + sizeof(UInt32) * 3, &uid, sizeof(UInt32));

				MemHandleUnlock(newrecordH);
				MemHandleUnlock(srecordH);
				DmReleaseRecord(ScoreDB, newIndex, true);
				DmReleaseRecord(ScoreDB, sindex, true);
				srecord = NULL;
			} while((uid = crecord->next) != NoRecord);

			MemHandleUnlock(crecordH);
			DmReleaseRecord(CompDB, cindex, true);
		}
		compuid = 0;
	} else if(compuid) {
		err = DmFindRecordByID(CompDB, compuid, &cindex);
		if(err) return err;

		crecordH = DmGetRecord(CompDB, cindex);
		if(!crecordH) return DmGetLastErr();

		crecord = MemHandleLock(crecordH);
		if(!crecord) return 1;

		if((stguid = mrecord->StgRef) != NoRecord)
		{
			while(stguid != NoRecord)
			{
				err = DmFindRecordByID(StageDB, stguid, &stgidx);
				if(err) return err;
				stgrecH = DmQueryRecord(StageDB, stgidx);
				if(!stgrecH) return DmGetLastErr();
				stgrec = MemHandleLock(stgrecH);
				if(!stgrec) return 1;

				/* Add a score record for every stage here */
				score = MemPtrNew(sizeof(Score));
				if(!score) return 1;

				/* Set default values */
				score->As = 0;
				score->Cs = 0;
				score->Ds = 0;
				score->PTs = 0;
				score->Misses = 0;
				score->Procs = 0;
				score->Time = 0.0;
				score->Zeroed = 0;
				score->Scored = 0;
				score->prev = scoreuid;
				score->next = NoRecord;
				score->StgRef = stguid;
				score->CptRef = compuid;

				newIndex = dmMaxRecordIndex;
				newrecordH = DmNewRecord(ScoreDB, &newIndex, ScoreDBPackedSize());
				if(!newrecordH)
				{
					MemPtrFree(score);
					return DmGetLastErr();
				}
				newrecord = MemHandleLock(newrecordH);
				if(!newrecord)
				{
					MemPtrFree(score);
					return 1;
				}
				ScoreDBPackRecord(score, newrecord);
				MemPtrFree(score);

				err = DmRecordInfo(ScoreDB, newIndex, NULL, &uid, NULL);
				if(err) return err;

				if(scoreuid == NoRecord && crecord->ScrRef == NoRecord)
				{
					DmWrite(crecord, sizeof(UInt16) * 2 + sizeof(UInt32), &uid, sizeof(UInt32));
				} else if(srecord) {
					DmWrite(srecord, sizeof(UInt8) * 6 + sizeof(double) + sizeof(UInt16) + sizeof(UInt32) * 3, &uid, sizeof(UInt32));
					MemHandleUnlock(srecordH);
					DmReleaseRecord(ScoreDB, sindex, true);
				}

				sindex = newIndex;
				srecordH = newrecordH;
				srecord = newrecord;
				stguid = stgrec->next;
				scoreuid = uid;
				MemHandleUnlock(stgrecH);
			}
			MemHandleUnlock(srecordH);
			DmReleaseRecord(ScoreDB, sindex, true);
		} /* else there are no stages in the match, so add no scores. */
		MemHandleUnlock(crecordH);
		DmReleaseRecord(CompDB, cindex, true);
	}
	MemHandleUnlock(mrecordH);

	return errNone;
}

static UInt16 ScoreDBGetRecordInList(UInt16 mindex, Int16 compsel, Int16 stagesel)
{
	MemHandle recordH;
	ScorePktPtr srecord = NULL; /* make gcc quiet */
	CompPktPtr crecord;
	UInt16 sindex, cindex;
	UInt32 scoreuid;
	Err err;

	if(stagesel == NoRecord) return NoRecord;

	if(compsel == NoRecord) return NoRecord;

	cindex = CompDBGetRecordInList(mindex, compsel);

	recordH = DmQueryRecord(CompDB, cindex);
	if(!recordH) return NoRecord;
	crecord = MemHandleLock(recordH);
	if(!crecord) return NoRecord;

	scoreuid = crecord->ScrRef;

	MemHandleUnlock(recordH);

	for(;stagesel > 0; stagesel--)
	{
		err = DmFindRecordByID(ScoreDB, scoreuid, &sindex);
		if(err) return NoRecord;

		recordH = DmQueryRecord(ScoreDB, sindex);
		if(!recordH) return NoRecord;
		srecord = MemHandleLock(recordH);
		if(!srecord) return NoRecord;

		scoreuid = srecord->next;

		MemHandleUnlock(recordH);
	}

	err = DmFindRecordByID(ScoreDB, scoreuid, &sindex);
	if(err) return NoRecord;

	return sindex;
}

Err ScoreDBGetScore(UInt16 match, Int16 stage, Int16 comp, ScorePtr score)
{
	ScorePktPtr packed;
	MemHandle recordH;
	UInt16 index;

	index = ScoreDBGetRecordInList(match, comp, stage);
	
	recordH = DmQueryRecord(ScoreDB, index);
	if(!recordH)
	{
		score = NULL;
		return DmGetLastErr();
	}
	packed = MemHandleLock(recordH);
	if(!packed)
	{
		score = NULL;
		return 1;
	}

	ScoreDBUnpackRecord(packed, score);
	MemHandleUnlock(recordH);

	return errNone;
}

Err ScoreDBUpdateScore(ScorePtr unpacked, UInt16 match, Int16 comp, Int16 stage)
{
	MemHandle recordH, oldH;
	ScorePktPtr recordP;
	UInt16 index;
	Err err;

	index = ScoreDBGetRecordInList(match, comp, stage);

	oldH = (MemHandle)true;

	recordH = DmNewHandle(ScoreDB, ScoreDBPackedSize());
	if(!recordH) return DmGetLastErr();

	recordP = MemHandleLock(recordH);
	if(!recordP) return 1;

	ScoreDBPackRecord(unpacked, recordP);
	MemHandleUnlock(recordH);

	err = DmAttachRecord(ScoreDB, &index, recordH, &oldH);
	if(err) return err;
	MemHandleFree(oldH);

	SetDirty(ScoreDB, &index);

	return errNone;
}

static Err ScoreDBDeleteRecord(UInt16 index)
{
	Err err;
	MemHandle recordH;
	ScorePktPtr score;
	CompPktPtr comp;
	UInt32 nextuid, prevuid, compuid = NoRecord;
	UInt16 next = NoRecord, prev = NoRecord, compindex = NoRecord;

	recordH = DmQueryRecord(ScoreDB, index);
	if(!recordH) return DmGetLastErr();
	score = MemHandleLock(recordH);
	if(!score) return 1;

	nextuid = score->next;
	prevuid = score->prev;
	compuid = score->CptRef;

	err = DmFindRecordByID(CompDB, compuid, &compindex);
	if(err) return err;

	MemHandleUnlock(recordH);

	if(prevuid != NoRecord)
	{
		err = DmFindRecordByID(ScoreDB, prevuid, &prev);
		if(err) return err;

		recordH = DmGetRecord(ScoreDB, prev);
		if(!recordH) return DmGetLastErr();
		score = MemHandleLock(recordH);
		if(!score) return 1;

		DmWrite(score, sizeof(UInt8) * 6 + sizeof(double) + sizeof(UInt16) + sizeof(UInt32) * 3, &nextuid, sizeof(UInt32));
		MemHandleUnlock(recordH);
		err = DmReleaseRecord(ScoreDB, prev, true);
		if(err) return err;
	} else {
		recordH = DmGetRecord(CompDB, compindex);
		if(!recordH) return DmGetLastErr();
		comp = MemHandleLock(recordH);
		if(!comp) return 1;

		err = DmWrite(comp, sizeof(UInt16) * 2 + sizeof(UInt32), &nextuid, sizeof(UInt32));
		if(err) return err;

		MemHandleUnlock(recordH);
		err = DmReleaseRecord(CompDB, compindex, true);
		if(err) return err;
	}

	if(nextuid != NoRecord)
	{
		err = DmFindRecordByID(ScoreDB, nextuid, &next);
		if(err) return err;

		recordH = DmGetRecord(ScoreDB, next);
		if(!recordH) return DmGetLastErr();
		score = MemHandleLock(recordH);
		if(!score) return 1;

		err = DmWrite(score, sizeof(UInt8) * 6 + sizeof(double) + sizeof(UInt16) + sizeof(UInt32) * 2, &prevuid, sizeof(UInt32));
		if(err) return err;

		MemHandleUnlock(recordH);
		err = DmReleaseRecord(ScoreDB, next, true);
		if(err) return err;
	}

	err = DmRemoveRecord(ScoreDB, index);
	if(err) return err;

	return errNone;
}

static Err ScoreDBDeleteScore(UInt16 match, Int16 compsel, Int16 stagesel)
{
	return ScoreDBDeleteRecord(ScoreDBGetRecordInList(match, compsel, stagesel));
}

void ResetCounters()
{
	numMatches = 0;
	numStagesInM = 0;
	numCompsInM = 0;
}

UInt16 MatchDBNumRecords()
{
	/*
	 * NOTE dmNumRecords() also counts deleted records. it's ok
	 * as long as we don't have any. then use
	 * dmNumRecordsInCategory(MatchDB, dmAllCategories)
	 */
	if(!numMatches) numMatches = DmNumRecords(MatchDB);

	return numMatches;
}

UInt32 GetMatchUID(UInt16 index)
{
	Err err;
	UInt32 ret = true;

	err = DmRecordInfo(MatchDB, index, NULL, &ret, NULL);
	if(err) return NoRecord;

	return ret;
}

UInt16 StageInMatchDBNumRecords(UInt16 index)
{
	MemHandle recordH = NULL;
	MatchPktPtr mrecord;

	if(index == NoRecord) return 0;

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH) return 0;

	mrecord = MemHandleLock(recordH);
	if(!recordH) return 0;

	numStagesInM = mrecord->StgCnt;

	MemHandleUnlock(recordH);

	return numStagesInM;
}

UInt16 CompInMatchDBNumRecords(UInt16 index)
{
	MemHandle recordH = NULL;
	MatchPktPtr mrecord;

	if(index == NoRecord) return 0;

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH) return 0;

	mrecord = MemHandleLock(recordH);
	if(!recordH) return 0;

	numCompsInM = mrecord->CptCnt;

	MemHandleUnlock(recordH);

	return numCompsInM;
}

UInt32 StageDBGetLastUID(UInt16 index)
{
	MemHandle recordH;
	MatchPktPtr mrecord;
	StagePktPtr srecord;
	UInt32 uid;

	if(index == NoRecord) return NoRecord;

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH) return NoRecord;

	mrecord = MemHandleLock(recordH);
	if(!recordH) return NoRecord;

	if((uid = mrecord->StgRef) != NoRecord)
	{
		do
		{
			/*
			 * unlock record from previous iteration.
			 */
			MemHandleUnlock(recordH);
			DmFindRecordByID(StageDB, uid, &index);

			recordH = DmQueryRecord(StageDB, index);
			if(!recordH) return NoRecord;

			srecord = MemHandleLock(recordH);
			if(!srecord) return NoRecord;

			if(srecord->next != NoRecord)
				uid = srecord->next;
			else
				break;

		} while(true);

		MemHandleUnlock(recordH);

		return uid;
	} 

	MemHandleUnlock(recordH);

	return NoRecord;
}

UInt32 CompDBGetLastUID(UInt16 index)
{
	MemHandle recordH;
	MatchPktPtr mrecord;
	CompPktPtr crecord;
	UInt32 uid;

	if(index == NoRecord) return NoRecord;

	recordH = DmQueryRecord(MatchDB, index);
	if(!recordH) return NoRecord;

	mrecord = MemHandleLock(recordH);
	if(!mrecord) return NoRecord;

	if((uid = mrecord->CptRef) != NoRecord)
	{
		do
		{
			/*
			 * unlock record from previous iteration.
			 */
			MemHandleUnlock(recordH);
			DmFindRecordByID(CompDB, uid, &index);

			recordH = DmQueryRecord(CompDB, index);
			if(!recordH) return NoRecord;

			crecord = MemHandleLock(recordH);
			if(!crecord) return NoRecord;

			if(crecord->next != NoRecord)
				uid = crecord->next;
			else
				break;

		} while(true);

		MemHandleUnlock(recordH);

		return uid;
	} 

	MemHandleUnlock(recordH);

	return NoRecord;
}

Err ResultsTableDataFree(UInt16 mindex, ResultTableDataPtr **DataArray)
{
	Int16 i;

        if(*DataArray)
        {
                if((i = CompInMatchDBNumRecords(mindex)) > 0)
                {
                        while(i)
                        {
                                i--;
                                MemHandleUnlock((*DataArray)[i]->CompDBrH);
				MemPtrFree((*DataArray)[i]);
                        }

                        MemPtrFree(*DataArray);
                }
        }

        return errNone;
}

Err ResultsTableDataQuery(UInt16 mindex, ResultTableDataPtr **DataArray)
{
	MemHandle recordH;
	CompPktPtr crecord = NULL; /* make gcc quiet */
	MatchPktPtr mrecord;
	UInt16 cindex, i;
	UInt32 compuid;
	Err err;

	recordH = DmQueryRecord(MatchDB, mindex);
	if(!recordH)
	{
		err = DmGetLastErr();
		return err;
	}

	mrecord = MemHandleLock(recordH);
	if(!mrecord) return 1;

	compuid = mrecord->CptRef;

	MemHandleUnlock(recordH);

	if((i = CompInMatchDBNumRecords(mindex)) > 0)
	{
		*DataArray = MemPtrNew(sizeof(ResultTableDataPtr) * i);
		if(!*DataArray) return -1;

		i = 0;
		do
		{
			err = DmFindRecordByID(CompDB, compuid, &cindex);
			if(err) return err;

			recordH = DmQueryRecord(CompDB, cindex);
			if(!recordH)
			{
				err = DmGetLastErr();
				return err;
			}

			crecord = MemHandleLock(recordH);
			if(!crecord) return 1;

			(*DataArray)[i] = MemPtrNew(sizeof(ResultTableData));
			if(!(*DataArray)[i]) return -1;

			(*DataArray)[i]->position = -1;
			(*DataArray)[i]->name = &(crecord->name);
			(*DataArray)[i]->percent = 0.0;
			(*DataArray)[i]->matchpoints = 0.0;
			(*DataArray)[i]->CompDBrH = recordH;

			compuid = crecord->next;
			i++;
		} while(compuid != NoRecord);
	} else {
		*DataArray = NULL;
	}

	return errNone;
}

Err ScoreTableDataFree(UInt16 mindex, ScoreTableDataPtr **DataArray)
{
	Int16 i;

	if(*DataArray)
	{
		if((i = CompInMatchDBNumRecords(mindex)) > 0)
		{
			while(i)
			{
				i--;
				MemPtrFree((*DataArray)[i]);
			}

			MemPtrFree(*DataArray);
		}
	}

        return errNone;
}

Err ScoreTableDataQuery(UInt16 mindex, Int16 stage, ScoreTableDataPtr **DataArray)
{
	MemHandle recordH;
	UInt16 i, numcomps, index;
	Err err;

	if((numcomps = CompInMatchDBNumRecords(mindex)) > 0)
	{
		*DataArray = MemPtrNew(sizeof(ScoreTableDataPtr) * numcomps);
		if(!*DataArray) return -1;

		for(i = 0; i < numcomps; i++)
		{
			index = ScoreDBGetRecordInList(mindex, i, stage);
			recordH = DmQueryRecord(ScoreDB, index);
			if(!recordH)
			{
				err = DmGetLastErr();
				MemPtrFree(*DataArray);
				*DataArray = NULL;
				return err;
			}

			(*DataArray)[i] = MemPtrNew(sizeof(ScoreTableData));
			if(!(*DataArray)[i]) return -1;

			(*DataArray)[i]->HF = 0.0;
			(*DataArray)[i]->comp = i;
			(*DataArray)[i]->ScoreDBrH = recordH;
		}

	} else {
		*DataArray = NULL;
	}

	return errNone;
}
